import datetime
import fnmatch
import mimetypes
import os
import pathlib
from dataclasses import dataclass, replace
from dataclasses import field
from functools import cached_property
from typing import Callable
from typing import Iterable
from typing import Optional
from typing import TypeVar
from typing import Union

try:
    from os import PathLike
except ImportError:  # Python 3.5 support
    from pathlib import PurePath as PathLike

PATH_DATA_CLASS = TypeVar('PATH_DATA_CLASS', bound='PathData')
DEFAULT_FILTERS = [
    lambda obj: not obj.is_hidden
]


@dataclass(order=True)
class PathData(PathLike):
    """
    Object representing all data related with data.
    """

    file_path: Union[str, PathLike] = field(compare=True)
    file_name: Optional[str] = field(compare=False, default=None)
    patterns: Union[str, Iterable[str]] = field(compare=False, default="*")
    dirs: bool = field(compare=False, default=True)
    dirs_only: bool = field(compare=False, default=False)
    recursive: bool = field(compare=False, default=False)
    depth: int = field(compare=False, default=0)
    filters: Union[str, bool, Iterable[Union[Callable[[PATH_DATA_CLASS], bool], str]]] = field(
        compare=False,
        default=False
    )

    def __post_init__(self):

        if isinstance(self.file_path, str):
            self.file_path = os.path.expandvars(self.file_path)

        while self.depth:
            self.file_path = os.path.dirname(self.file_path)
            self.depth -= 1

        # is only one object for each system file.
        self.file_path = self.abspath

        self.parse_filters()

    def parse_filters(self):
        def __parse_filter(filter_data):
            if isinstance(filter_data, str):
                if filter_data[0] == '!':
                    default = False
                    filter_data = filter_data[1:]
                else:
                    default = True
                return lambda obj: default if getattr(obj, filter_data) else not default
            return filter_data
        if isinstance(self.filters, str):
            self.filters = [self.filters]

        elif isinstance(self.filters, bool):
            self.filters = DEFAULT_FILTERS if self.filters else ()

        self.filters = list(map(__parse_filter, self.filters))

    @classmethod
    def home(cls, *args, **kwargs):
        file_path = pathlib.Path.home()
        return cls(file_path, *args, **kwargs)

    @classmethod
    def root(cls, *args, **kwargs):
        file_path = pathlib.Path("/")
        return cls(file_path, *args, **kwargs)

    @classmethod
    def main(cls, *paths, **kwargs):
        file_path = os.getcwd()
        __main = cls(file_path, **kwargs)
        if paths:
            return __main(*paths)
        else:
            return __main

    @cached_property
    def parent(self):
        file_path = os.path.dirname(self)
        return replace(self, file_path=file_path)

    def __call__(self, *paths: Union[str, Iterable[str]], **options: dict) -> PATH_DATA_CLASS:
        """
        paths: one file path or list of file path's elements
        Return absolute path of give file in context of base parent.
        """
        if len(paths) == 1 and paths[0] and len(paths[0]):
            if isinstance(paths[0], str):
                "if is one argument, and it is a string, get it as file_name"
                file_name = paths[0]
                if file_name[0] != '/':
                    file_name = self.join(file_name)
            else:
                "if is one argument, and it is not a string, get it as argument list to join"
                file_name = self.join(paths[0])
        else:
            file_name = self.join(*paths)

        options["file_name"] = options.get("file_name")

        return replace(self, file_path=file_name, **options)

    def __fspath__(self) -> str:
        """for PathLike interface"""
        return self.abspath

    def __hash__(self) -> int:
        """For order"""
        return hash(self.__fspath__())

    def __eq__(self, other):
        return self.__fspath__() == getattr(other, "__fspath__", lambda: other)()

    def __str__(self):
        return self.abspath

    def __iter__(self) -> Iterable[PATH_DATA_CLASS]:
        return self.iterator()

    @cached_property
    def path(self) -> pathlib.Path:
        """Always return path object"""
        return pathlib.Path(self.file_path)

    @cached_property
    def abspath(self) -> str:
        """
        Return absolute path of base directory
        """
        return os.path.abspath(self.file_path)

    @cached_property
    def name(self) -> str:
        return str(self.file_name or self.path.name)

    @cached_property
    def is_dir(self) -> bool:
        return self.path.is_dir()

    @cached_property
    def is_hidden(self) -> bool:
        return self.name[0] == '.'

    @cached_property
    def created_at(self) -> Optional[datetime.datetime]:
        system_time = self.path.stat().st_ctime
        return datetime.datetime.fromtimestamp(system_time)

    @cached_property
    def change_at(self) -> Optional[datetime.datetime]:
        system_time = self.path.stat().st_mtime
        return datetime.datetime.fromtimestamp(system_time)

    @cached_property
    def size(self) -> int:
        file_size = os.path.getsize(self)
        return file_size

    @cached_property
    def mime_type(self):
        return mimetypes.guess_type(self.path)[0]

    def join(self, *paths) -> str:

        if self.is_dir:
            __path = self.path
        else:
            "if is a file, join to parent"
            __path = self.parent.path

        return os.path.join(__path, *paths)

    @cached_property
    def check(self):
        return all(map(lambda filter_: filter_(self), self.filters))

    def iterator(self, patterns: Union[str, Iterable[str]] = None, **options) -> Iterable[PATH_DATA_CLASS]:
        """
        :param patterns: patterns as a list or in string separate by ';'
        :yield: PathData object for each file

        Return files in given directory where names pass one of given patterns
        as a default return all files pass one from pattern list.
        """

        assert self.is_dir

        patterns = patterns or self.patterns

        if isinstance(patterns, str):
            patterns = patterns.split(';')

        for path in self.all(**options):
            name = path.name
            for pattern in patterns:
                if fnmatch.fnmatch(name, pattern):
                    obj = replace(self, file_path=path)
                    if obj.check:
                        yield obj
                    break

    def all(
            self,
            dirs: bool = None,
            dirs_only: bool = None,
            recursive: bool = None,
    ) -> Iterable[PathLike]:
        """
        :param recursive: walk recursively through subdirectories
        :param dirs: yield directories too
        :param dirs_only: yield only directories
        :yield: name of each file

        Return directory names.
        """

        assert self.is_dir

        dirs = self.dirs if dirs is None else dirs
        dirs_only = self.dirs_only if dirs_only is None else dirs_only
        recursive = self.recursive if recursive is None else recursive

        dirs_list: list[PATH_DATA_CLASS] = [self]

        while dirs_list:
            dir_ = dirs_list.pop()
            assert dir_.is_dir
            for path in dir_.path.iterdir():
                if path.is_dir():
                    if recursive:
                        dirs_list.append(replace(dir_, file_path=path))
                    if not dirs:
                        continue
                else:
                    if dirs_only:
                        continue
                yield path

    def ls(self, patterns: Union[str, Iterable[str]] = None, **options) -> Optional[list[PATH_DATA_CLASS]]:
        return list(self.iterator(patterns, **options)) if self.is_dir else None

    def __len__(self):
        if not self.is_dir:
            return None
        assert self.is_dir
        return len(self.ls())



import os
import sys
import re

try:
    from os import PathLike
except ImportError:  # Python 3.5 support
    from pathlib import PurePath as PathLike

Openable = (str, PathLike)

from abc import abstractmethod
from typing import Optional, Iterable, Type, AnyStr
from logging import getLogger

from .mapping import MappingConst

logger = getLogger(__name__)

class Value:

    def __call__(self, *envs: dict) -> Optional[AnyStr]:
        for env in envs:
            result = self.get_value(env)
            if result is not None:
                return result
        return None

    @abstractmethod
    def get_value(self, config: object) -> AnyStr:
        pass


class EnvValue(Value):
    def __init__(self, var_name: str):
        self.var_name = var_name

    def get_value(self, env):
        return env.get(self.var_name)


def join_path(*args):
    return os.path.join(*args)


def join_path_value(dir_id):
    """
    produce a special parser with join another variable value to given path
    """
    class JoinPathValue(EnvValue):
        def get_value(self, env):
            if dir_id in env:
                value = join_path(env[dir_id], self.var_name)
                return value
    return JoinPathValue


def format_string(string):

    class FormatString(EnvValue):
        def get_value(self, env):
            return string.format(env)

    return FormatString

def bool_value(value):
    try:
        i = float(value)
        return bool(i)
    except ValueError:
        pass

    s = str(value).lower()
    if "true" in s:
        return True
    if "false" in s:
        return False
    return None


def expand_defaults(init):
    return (
        map(
            lambda args:
            (lambda key, value=None, parser=str, env_key=None, parse_default=True:
             (
                 key,
                 value,
                 parser,
                 key if env_key is None else env_key,
                 parse_default
             )
             )(*args),
            init
        )
    )


class Config(MappingConst):

    @abstractmethod
    def get_values(self) -> dict:
        return {}

    def __init__(self):
        super().__init__()
        self.__dict__ = self.get_values()


def init_envs(envs: Optional[Iterable]) -> list:
    return list(envs) if envs is not None else []


class ConfigEnv(Config):
    DEFAULTS = []

    def get_envs(self, envs: Optional[Iterable[dict]] = None) -> list:
        envs = init_envs(envs)
        envs.append(os.environ)
        return envs

    def get_values(self, init=None):
        if init is not None:
            defaults = expand_defaults(init)
        else:
            defaults = self.DEFAULTS
        values = {}
        envs = self.get_envs()
        for key, value, parser, env_key, parse_default in defaults:
            env_val = EnvValue(env_key)(*envs)

            if env_val is not None:
                value = parser(env_val)
            elif parse_default and not callable(value):
                value = parser(value)

            while callable(value):
                value = value(values, *envs)
                if parser:
                    value = parser(value)
                    parser = False

            values[key] = value
        return values

    def update(self, init):
        return super().update(self.get_values(init))


class ConfigEnvFiles(ConfigEnv):
    ENV_FILES = []

    def get_env_files(self) -> dict:
        env = {}
        for env_file_path in self.ENV_FILES:
            env = read_dot_env(env_file_path, **env)
        return env

    def get_envs(self, envs: Optional[Iterable[dict]] = None) -> list:
        envs = super().get_envs(envs)
        envs.append(self.get_env_files())
        return envs


def read_dot_env(env_file=None, **env):
    """
    From: https://github.com/joke2k/django-environ

    Read a .env file into os.environ.

    If not given a path to a dotenv path, does filthy magic stack
    backtracking to find the dotenv in the same directory as the file that
    called read_env.

    Existing environment variables take precedent and are NOT overwritten
    by the file content. ``overwrite=True`` will force an overwrite of
    existing environment variables.

    Refs:
    - https://wellfire.co/learn/easier-12-factor-django
    - https://gist.github.com/bennylope/2999704

    :param env_file: The path to the `.env` file your application should
        use. If a path is not provided, `read_env` will attempt to import
        the Django settings module from the Django project root.
    :param **env: Any additional keyword arguments provided directly
        to read_env will be added to the environment. If the key matches an
        existing environment variable, the value will be overridden.
    """
    if env_file is None:
        frame = sys._getframe()
        env_file = os.path.join(
            os.path.dirname(frame.f_back.f_code.co_filename),
            '.env'
        )
        if not os.path.exists(env_file):
            logger.info(
                "%s doesn't exist - if you're not configuring your "
                "environment separately, create one." % env_file)
            return

    try:
        if isinstance(env_file, Openable):
            # Python 3.5 support (wrap path with str).
            with open(str(env_file)) as f:
                content = f.read()
        else:
            with env_file as f:
                content = f.read()
    except OSError:
        logger.info(
            "%s not found - if you're not configuring your "
            "environment separately, check this." % env_file)
        return

    logger.debug('Read environment variables from: {}'.format(env_file))

    def _keep_escaped_format_characters(match):
        """Keep escaped newline/tabs in quoted strings"""
        escaped_char = match.group(1)
        if escaped_char in 'rnt':
            return '\\' + escaped_char
        return escaped_char

    for line in content.splitlines():
        m1 = re.match(r'\A(?:export )?([A-Za-z_0-9]+)=(.*)\Z', line)
        if m1:
            key, val = m1.group(1), m1.group(2)
            m2 = re.match(r"\A'(.*)'\Z", val)
            if m2:
                val = m2.group(1)
            m3 = re.match(r'\A"(.*)"\Z', val)
            if m3:
                val = re.sub(r'\\(.)', _keep_escaped_format_characters,
                             m3.group(1))
            env[key] = str(val)
        elif not line or line.startswith('#'):
            # ignore warnings for empty line-breaks or comments
            pass
        else:
            logger.warning('Invalid line: %s', line)

    return env


def get_config(
        defaults: Iterable[tuple],
        env_files: Optional[Iterable[AnyStr]] = None,
        config_class: Optional[Type[Config]] = ConfigEnvFiles
) -> Type[Config]:

    class ConfigClass(config_class):
        DEFAULTS = expand_defaults(defaults)
        ENV_FILES = env_files or []

    return ConfigClass



import os
from dataclasses import dataclass
from typing import Iterable
from typing import TypeVar
from typing import Union
from .path_data import PathData


def all_files(root: str, patterns: str = "*", single_level: bool = False, yield_folders: bool = False) -> Iterable[str]:
    """return all files in given directory"""
    root_dir = PathGenerator(root)
    return root_dir.ls(
            patterns=patterns,
            recursive=not single_level,
            dirs=yield_folders
        )


DIR_CLASS = TypeVar('DIR_CLASS', bound='Dir')


@dataclass
class PathGenerator(PathData):
    """
    Special version of PathData, for configuration tasks.
    As a function return a string value of unique absolute paths of file.
    As a iterator returns a string value of absolute paths of contained files.
    """

    dirs: bool = False
    dirs_only: bool = False
    recursive: bool = True

    def __call__(self, *args, **kwargs) -> str:
        return str(self.get_dir(*args, **kwargs))

    def ls(self, patterns: Union[str, Iterable[str]] = None, **options) -> Iterable[str]:
        iterator = super().iterator(patterns, **options)
        if not iterator:
            return iterator
        return [str(obj) for obj in iterator]

    def get_dir(self, *args, **kwargs) -> DIR_CLASS:
        return super().__call__(*args, **kwargs)


def get_dir(file=None, *args, **kwargs):
    if file is None:
        return PathGenerator.main(*args, **kwargs)
    else:
        return PathGenerator(file, *args, **kwargs)


def get_home():
    return PathGenerator.home()
